    # -*- coding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (C) 2004-2010, 2013, 2014 Tiny SPRL (<http://tiny.be>).
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import time
import simplejson
from lxml import etree

from openerp.osv import fields, osv
from openerp.osv.orm import Model
from openerp.tools import DEFAULT_SERVER_DATE_FORMAT as DF
from openerp.tools.translate import _
from openerp import netsvc
import openerp.addons.decimal_precision as dp


class sale_order(Model):
    _name = 'sale.order'
    _inherit = 'sale.order'
    _columns = {
        'date_order':
            fields.date('Start Date', required=True, readonly=True,
                        select=True,
                        states={'draft': [('readonly', False)],
                                'sent': [('readonly', False)]}),
        'end_date':
            fields.date('End Date'),
        'order_line':
            fields.one2many('sale.order.line', 'order_id', 'Order Lines',
                            readonly=True,
                            states={'draft': [('readonly', False)],
                                    'sent': [('readonly', False)],
                                    'manual': [('readonly', False)]}),
        'contact_ids':
            fields.many2many('res.partner', 'sale_order_partner_rel',
                             'order_id', 'partner_id', 'Contacts'),
        'month_eur_rate':
            fields.float('Month Rate (EUR - USD)', digits=(12, 4))
    }

    def _get_month_rate(self, cr, uid, context=None):
        res = False
        currency = self.pool.get('res.currency')
        eur_ids = currency.search(cr, uid, [('name', '=', 'EUR')], context=context)
        if eur_ids:
            ctx = context.copy()
            ctx.update({'currency_rate_type_id': 1})
            sale_rate = currency.browse(cr, uid, eur_ids[0], context=context).rate
            res = 1 / sale_rate
        return res

    _defaults = {
        'shop_id': 1,
        'month_eur_rate': _get_month_rate
    }

    _order = 'create_date desc'

    def write(self, cr, uid, ids, vals, context=None):
        res = super(sale_order, self).write(cr, uid, ids, vals, context)
        if 'state' in vals and vals['state'] in ['confirmed', 'done', 'cancel']:
            sol = self.pool.get('sale.order.line')
            sol_ids = []
            for s in self.browse(cr, uid, ids, context):
                sol_ids += [sl.id for sl in s.order_line]
            if sol_ids:
                sol.write(cr, uid, sol_ids, {'state': vals['state']}, context)
        return res

    def reset_workflow(self, cr, uid, ids, context=None):
        order = self.browse(cr, uid, ids[0], context)
        if order.state in ['draft', 'manual', 'sent']:
            self.write(cr, uid, [order.id], {'state': 'draft'})
            wf_service = netsvc.LocalService("workflow")
            wf_service.trg_delete(uid, 'sale.order', order.id, cr)
            wf_service.trg_create(uid, 'sale.order', order.id, cr)
        return True


class sale_order_line(Model):
    _name = 'sale.order.line'
    _inherit = 'sale.order.line'
    _inherits = {'travel.line': 'travel_line_id'}

    def _get_paxs(self, cr, uid, ids, fields, args, context=None):
        result = {}
        for obj in self.browse(cr, uid, ids, context):
            result[obj.id] = obj.adults + obj.children
        return result

    _columns = {
        'travel_line_id': fields.many2one('travel.line', 'Travel Line',
                                          ondelete="cascade", select=True),
        'adults': fields.integer('Adults'),
        'children': fields.integer('Children'),
        'supplier_id': fields.many2one('res.partner', 'Supplier',
                                       domain="[('supplier', '=', True)]",
                                       readonly=True,
                                       states={'draft': [('readonly', False)],
                                               'cancel': [('readonly', False)]}),
        'reservation_number': fields.char('Reservation', size=64),
        'res_pax_ids': fields.many2many('res.pax',
                                        'sale_order_line_res_pax_related',
                                        'line_id', 'pax_id', 'Paxs'),
        'customer_id': fields.related('order_id', 'partner_id',
                                      type='many2one', relation='res.partner',
                                      readonly=True, string='Customer',
                                      store=True),
        'order_start_date': fields.related('order_id', 'date_order',
                                           type='date', string='In',
                                           store=True),
        'order_end_date': fields.related('order_id', 'end_date', type='date',
                                         string='Out', store=True),
        'description': fields.text('Description'),
        'paxs': fields.function(_get_paxs, type='integer', method=True,
                                string='Paxs', store=True),
        'price_unit_cost': fields.float('Cost Price',
                            digits_compute=dp.get_precision('Product Price'),
                            required=True),
        'cost_currency_id': fields.many2one('res.currency', 'Cost')
    }

    def show_product(self, cr, uid, ids, context=None):
        line = self.browse(cr, uid, ids[0], context)
        categ = line.product_type
        return {
            'type': 'ir.actions.act_window',
            'res_model': 'product.' + categ,
            'res_id': line.travel_product_id.id,
            'view_mode': 'form',
            'target': 'new'
        }

    def to_confirm(self, cr, uid, ids, context=None):
        res = self.write(cr, uid, ids, {'state': 'confirmed'}, context)
        return res

    def to_cancel(self, cr, uid, ids, context=None):
        res = self.write(cr, uid, ids, {'state': 'cancel'}, context)
        return res

    def to_send(self, cr, uid, ids, context=None):
        assert len(ids) == 1, 'This option should only be used for a single id at a time.'
        imd = self.pool.get('ir.model.data')
        template_id = False
        try:
            compose_form_id = imd.get_object_reference(cr, uid, 'mail',
                                        'email_compose_message_wizard_form')[1]
        except ValueError:
            compose_form_id = False
        ctx = dict(context)
        ctx.update({
            'default_model': 'sale.order.line',
            'default_res_id': ids[0],
            'default_use_template': False,
            'default_template_id': template_id,
            'default_composition_mode': 'comment'
        })
        return {
            'type': 'ir.actions.act_window',
            'view_type': 'form',
            'view_mode': 'form',
            'res_model': 'mail.compose.message',
            'views': [(compose_form_id, 'form')],
            'view_id': compose_form_id,
            'target': 'new',
            'context': ctx,
        }

    def get_supplier(self, line):
        if line.supplier_id:
            return line.supplier_id
        elif line.product_id.seller_id:
            return line.product_id.seller_id
        else:
            raise osv.except_osv(_('Error!'),
                  _("There is at least one sale order line without supplier"))

    def get_supplier_info(self, cr, uid, line, context=None):
        supplierinfo_obj = self.pool.get('product.supplierinfo')
        supplier = self.get_supplier(line)
        to_search = [
            ('name', '=', supplier.id),
            ('product_id', '=', line.travel_product_id.product_tmpl_id.id)
        ]
        sinfo = supplierinfo_obj.search(cr, uid, to_search, context=context)
        return sinfo and sinfo[0] or False

    def onchange_product(self, cr, uid, ids, product, context=None):
        res = {'value': {'product_type': False}}
        if product:
            pmodel, ptid = product.split(',')
            pid = self.pool.get(pmodel).read(cr, uid, int(ptid), ['product_id'],
                                             context)['product_id']
            res['value'] = {
                'product_id': pid,
                'product_type': pmodel.split('.')[1]
            }
        return res

    def product_id_change(self, cr, uid, ids, pricelist, product, qty=0,
            uom=False, qty_uos=0, uos=False, name='', partner_id=False,
            lang=False, update_tax=True, date_order=False, packaging=False,
            fiscal_position=False, flag=False, context=None):
        context = context or {}
        lang = lang or context.get('lang', False)
        if not  partner_id:
            raise osv.except_osv(_('No Customer Defined !'),
                                 _('Before choosing a product,\n select a customer in the sales form.'))
        warning = {}
        product_uom_obj = self.pool.get('product.uom')
        partner_obj = self.pool.get('res.partner')
        product_obj = self.pool.get('product.product')
        supplier_id = context.get('supplier_id', False)
        params = context.get('params', {})
        context = {
            'lang': lang,
            'partner_id': partner_id,
            'supplier_id': supplier_id,
            'params': params
        }
        if partner_id:
            lang = partner_obj.browse(cr, uid, partner_id).lang
        context_partner = {'lang': lang, 'partner_id': partner_id}

        if not product:
            return {'value': {'th_weight': 0,
                'product_uos_qty': qty}, 'domain': {'product_uom': [],
                   'product_uos': []}}
        if not date_order:
            date_order = time.strftime(DF)

        result = {}
        warning_msgs = ''
        product_obj = product_obj.browse(cr, uid, product,
                                         context=context_partner)
        uom2 = False
        if uom:
            uom2 = product_uom_obj.browse(cr, uid, uom)
            if product_obj.uom_id.category_id.id != uom2.category_id.id:
                uom = False
        if uos:
            if product_obj.uos_id:
                uos2 = product_uom_obj.browse(cr, uid, uos)
                if product_obj.uos_id.category_id.id != uos2.category_id.id:
                    uos = False
            else:
                uos = False
        afp = self.pool.get('account.fiscal.position')
        fpos = fiscal_position and afp.browse(cr, uid, fiscal_position) or False
        if update_tax:
            result['tax_id'] = afp.map_tax(cr, uid, fpos, product_obj.taxes_id)
        if not flag:
            pp = self.pool.get('product.product')
            result['name'] = pp.name_get(cr, uid, [product_obj.id],
                                         context=context_partner)[0][1]
            if product_obj.description_sale:
                result['name'] += '\n' + product_obj.description_sale
        domain = {}
        if (not uom) and (not uos):
            result['product_uom'] = product_obj.uom_id.id
            if product_obj.uos_id:
                result['product_uos'] = product_obj.uos_id.id
                result['product_uos_qty'] = qty * product_obj.uos_coeff
                uos_category_id = product_obj.uos_id.category_id.id
            else:
                result['product_uos'] = False
                result['product_uos_qty'] = qty
                uos_category_id = False
            result['th_weight'] = qty * product_obj.weight
            domain = {'product_uom':
                      [('category_id', '=', product_obj.uom_id.category_id.id)],
                        'product_uos':
                      [('category_id', '=', uos_category_id)]}
        elif uos and not uom:
            result['product_uom'] = product_obj.uom_id and product_obj.uom_id.id
            result['product_uom_qty'] = qty_uos / product_obj.uos_coeff
            result['th_weight'] = result['product_uom_qty'] * product_obj.weight
        elif uom:
            default_uom = product_obj.uom_id and product_obj.uom_id.id
            q = product_uom_obj._compute_qty(cr, uid, uom, qty, default_uom)
            if product_obj.uos_id:
                result['product_uos'] = product_obj.uos_id.id
                result['product_uos_qty'] = qty * product_obj.uos_coeff
            else:
                result['product_uos'] = False
                result['product_uos_qty'] = qty
            result['th_weight'] = q * product_obj.weight

        if not uom2:
            uom2 = product_obj.uom_id

        if not pricelist:
            warn_msg = _('You have to select a pricelist or a customer in the sales form !\n'
                    'Please set one before choosing a product.')
            warning_msgs += _("No Pricelist ! : ") + warn_msg + "\n\n"
        else:
            price = self.pool.get('product.pricelist').price_get(cr, uid, [pricelist],
                    product, qty or 1.0, partner_id, {
                        'uom': uom or result.get('product_uom'),
                        'date': date_order,
                        'supplier_id': supplier_id,
                        'params': params
                        })[pricelist]
            if price is False:
                pass
                #warn_msg = _("Cannot find a pricelist line matching this product and quantity.\n"
                #        "You have to change either the product, the quantity or the pricelist.")
                #
                #warning_msgs += _("No valid pricelist line found ! :") + warn_msg +"\n\n"
            else:
                result.update({'price_unit': price})

            if supplier_id:
                sc = partner_obj.browse(cr, uid, supplier_id, context).ref
                if sc:
                    sc_id = self.pool.get('res.currency').search(cr, uid,
                                                        [('name', '=', sc)],
                                                        context=context)
                    if sc_id:
                        result.update({'cost_currency_id': sc_id[0]})
            cost_price = self.show_cost_price(cr, uid, result, product, qty,
                                              partner_id, uom, date_order,
                                              supplier_id, params, pricelist,
                                              context)
            result.update({'price_unit_cost': cost_price})

        if warning_msgs:
            warning = {
                       'title': _('Configuration Error!'),
                       'message': warning_msgs
                    }
        if not context.get('supplier_id', False):
            result.update({'supplier_id': product_obj.seller_id.id})
        return {'value': result, 'domain': domain, 'warning': warning}

    def show_cost_price(self, cr, uid, result, product, qty, partner_id, uom,
                        date_order, supplier_id, params, pricelist,
                        context=None):
        pl_obj = self.pool.get('product.pricelist')
        cp_ids = pl_obj.search(cr, uid, [('name', '=', 'Cost Pricelist')],
                               context=context)
        if cp_ids:
            cp_id = cp_ids[0]
            cost_price = pl_obj.price_get(cr, uid, [cp_id],
                            product, qty or 1.0, partner_id, {
                                'uom': uom or result.get('product_uom'),
                                'date': date_order,
                                'supplier_id': supplier_id,
                                'params': params
                                })[cp_id]
            sale_currency_id = pl_obj.browse(cr, uid, pricelist).currency_id.id
            cost_currency_id = pl_obj.browse(cr, uid, cp_id).currency_id.id
            if sale_currency_id != cost_currency_id:
                cr_obj = self.pool.get('res.currency')
                cost_price = cr_obj.compute(cr, uid, cost_currency_id,
                                            sale_currency_id, cost_price,
                                            round=False, context=context)
            return cost_price
        return 0.0

    def fields_view_get(self, cr, uid, view_id=None, view_type='form',
                        context=None, toolbar=False, submenu=False):
        if not context:
            context = {}
        res = super(sale_order_line, self).fields_view_get(cr, uid, view_id,
                view_type, context=context, toolbar=toolbar, submenu=submenu)
        if view_type == 'form':
            tl = self.pool.get('travel.line')
            new_fields = tl.get_travel_fields(cr, uid, context)
            res['fields'].update(new_fields)
            doc = etree.XML(res['arch'])
            parent_node = doc.xpath("//group[@name='extended']")
            if parent_node:
                parent_node = parent_node[0]
                sd = doc.xpath("//field[@name='start_date']")[0]
                ocg = sd.get('on_change', False)
                ctx = sd.get('context', False)
                ctx = tl.build_ctx(ctx, new_fields)
                # TODO: pasar la modificacion de los campos standards para un
                # metodo aparte
                sd.set('context', ctx)
                ed = doc.xpath("//field[@name='end_date']")[0]
                ed.set('context', ctx)
                ad = doc.xpath("//field[@name='adults']")[0]
                ad.set('context', ctx)
                ch = doc.xpath("//field[@name='children']")[0]
                ch.set('context', ctx)
                keys = new_fields.keys()
                keys.sort()
                for f in keys:
                    n = etree.Element("field")
                    n.set('name', f)
                    modifiers = tl.build_attr(f)
                    n.set('modifiers', simplejson.dumps(modifiers))
                    n.set('context', ctx)
                    n.set('on_change', ocg)
                    parent_node.append(n)
                res['arch'] = etree.tostring(doc)
        return res

    # def _get_cost_currency_id(self, cr, uid, context=None):
    #
    #     return False

    _defaults = {
        'start_date': lambda s, c, u, ctx: ctx.get('start', False),
        'end_date': lambda s, c, u, ctx: ctx.get('end', False),
        'state': 'draft',
        'adults': 1,
        'children': 0,
        # 'cost_currency_id': _get_cost_currency_id,
    }

    # def create(self, cr, uid, vals, context=None):
    #     if vals.get('cost_currency_id', False):
    #
    #     res_id = super(sale_order_line, self).create(cr, uid, vals, context)
    #     return res_id
    #
    # def write(self, cr, uid, ids, vals, context=None):
    #     if not context:
    #         context = {}
    #
    #     if context.get('flag_rate', False):
    #         rates = {}
    #         for obj in self.browse(cr, uid, ids, context):
    #             rates.update({obj.id: obj.rate})
    #         context.update({'rates': rates})
    #     res = super(res_currency_rate, self).write(cr, uid, ids, vals, context)
    #     self.update_rate(cr, uid, ids, context)
    #     return res

    def _check_cost(self, cr, uid, ids, context=None):
        obj = self.browse(cr, uid, ids[0], context=context)
        if obj.price_unit_cost <= 0.0 or obj.price_unit_cost > obj.price_unit:
            return False
        # if not obj.cost_currency_id.id:
        #     if obj.price_unit_cost <= 0.0 or obj.price_unit_cost > obj.price_unit:
        #         return False
        # else:
        #     sale_order = self.pool.get('sale.order')
        #     sale_order_obj = sale_order.browse(cr, uid, obj.order_id.id, context=context)
        #     pl_obj = self.pool.get('product.pricelist')
        #     sale_currency_id = pl_obj.browse(cr, uid, sale_order_obj.pricelist_id.id).currency_id.id
        #     if obj.cost_currency_id.id != sale_currency_id:
        #         cr_obj = self.pool.get('res.currency')
        #         cost_price = cr_obj.compute(cr, uid, obj.cost_currency_id.id,
        #                                     sale_currency_id, obj.price_unit_cost,
        #                                     round=False, context=context)
        #         if cost_price <= 0.0 or cost_price > obj.price_unit:
        #             return False
        #     else:
        #         if obj.price_unit_cost <= 0.0 or obj.price_unit_cost > obj.price_unit:
        #             return False
        return True

    _constraints = [
        (_check_cost, 'Cost Price (x) => 0 < x < Income',
                      ['price_unit_cost', 'price_unit']),
    ]
